/**
 * @file
 * VuoQTCapture implementation.
 *
 * @copyright Copyright © 2012–2014 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "module.h"
#include "/usr/include/time.h"
#include "VuoQTCapture.h"
#include "VuoQtListener.h"
#include "VuoTriggerSet.hh"
#include <QTKit/QTkit.h>

#ifdef VUO_COMPILER
VuoModuleMetadata({
					  "title" : "VuoQtCapture",
					  "dependencies" : [
						"VuoImage",
						"VuoVideoInputDevice",
						"VuoList_VuoVideoInputDevice",
						"QTKit.framework",
						"VuoQtListener"
					  ]
				 });
#endif

/**
 *	A singleton class that listens for changes in available video input devices.
 */
@interface OnDevicesChangedListener : NSObject
{
	VuoTriggerSet<VuoList_VuoVideoInputDevice> callbacks;							///< A TriggerSet of VuoList_VuoVideoInputDevice output triggers.
	VuoList_VuoVideoInputDevice devices;											///< An always up to date list of available input devices.
}

- (id) init;																		///< Init and register for NSNotification events.
- (void) devicesDidChange:(NSNotification*)notif;									///< Called when a change is detected.
- (void) refreshDevices;															///< Refresh the cache of available input devices.
- (void) addCallback:(void(*)(VuoList_VuoVideoInputDevice)) outputTrigger;			///< Add a callback to be triggered when the input list changes.
- (void) removeCallback:(void(*)(VuoList_VuoVideoInputDevice)) outputTrigger;		///< Remove a callback from the trigger set.
@end

@implementation OnDevicesChangedListener

- (id) init
{
	if(!self)
		self = [super init];

	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(devicesDidChange:) name:QTCaptureDeviceWasConnectedNotification object:nil];
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(devicesDidChange:) name:QTCaptureDeviceWasDisconnectedNotification object:nil];

	[self refreshDevices];

	return self;
}

- (void) addCallback:(void(*)(VuoList_VuoVideoInputDevice)) outputTrigger
{
	callbacks.addTrigger(outputTrigger);
	outputTrigger(devices);
}

- (void) removeCallback:(void(*)(VuoList_VuoVideoInputDevice)) outputTrigger
{
	callbacks.removeTrigger(outputTrigger);
}

- (void) devicesDidChange:(NSNotification*)notif
{
	[self refreshDevices];
	callbacks.fire(devices);
}

- (void) refreshDevices
{
	if(devices != NULL)
		VuoRelease(devices);

	devices = NULL;

	devices = VuoListCreate_VuoVideoInputDevice();
	VuoRegisterSingleton(devices);
	
	NSArray *inputDevices = [[[QTCaptureDevice inputDevicesWithMediaType:QTMediaTypeVideo] arrayByAddingObjectsFromArray:[QTCaptureDevice inputDevicesWithMediaType:QTMediaTypeMuxed]] retain];
	
	for(QTCaptureDevice *dev in inputDevices) {
		VuoListAppendValue_VuoVideoInputDevice( devices, VuoVideoInputDevice_make([[dev modelUniqueID] UTF8String], [[dev localizedDisplayName] UTF8String]) );
	}

	[inputDevices release];
}
@end

static OnDevicesChangedListener *deviceListener;	///< A static reference to the OnDevicesChangedListener singleton.

/**
 *	Remove an output trigger from the OnDevicesChangedListener TriggerSet.
 */
void VuoQTCapture_removeOnDevicesChangedCallback( VuoOutputTrigger(devicesDidChange, VuoList_VuoVideoInputDevice) )
{
	if(deviceListener != nil)	
		[deviceListener removeCallback:devicesDidChange];
}

/**
 *	Add an output trigger to the OnDevicesChangedListener TriggerSet to be fired when available input devices changes.
 */
void VuoQTCapture_addOnDevicesChangedCallback( VuoOutputTrigger(devicesDidChange, VuoList_VuoVideoInputDevice) )
{
	if(deviceListener == nil)
		deviceListener = [[OnDevicesChangedListener alloc] init];
	[deviceListener addCallback:devicesDidChange];
}

/**
 * Returns a list of all available video capture devices.
 */
VuoList_VuoVideoInputDevice VuoQTCapture_getInputDevices(void)
{
	NSArray *inputDevices = [[[QTCaptureDevice inputDevicesWithMediaType:QTMediaTypeVideo] arrayByAddingObjectsFromArray:[QTCaptureDevice inputDevicesWithMediaType:QTMediaTypeMuxed]] retain];
	
	VuoList_VuoVideoInputDevice devices = VuoListCreate_VuoVideoInputDevice();

	for(QTCaptureDevice *dev in inputDevices) {
		VuoListAppendValue_VuoVideoInputDevice( devices, VuoVideoInputDevice_make([[dev modelUniqueID] UTF8String], [[dev localizedDisplayName] UTF8String]) );
	}

	[inputDevices release];
		
	return devices;
}

/**
 *	Forward declaration of destructor method.
 */
void VuoQTCapture_free(VuoQTCapture movie);

/**
 * Creates a new VuoQTCapture instance for receiving video frames from `inputDevice`.
 */
VuoQTCapture VuoQTCapture_make(VuoVideoInputDevice inputDevice, VuoOutputTrigger(receivedFrame, VuoVideoFrame))
{
	VuoQtListener *listener = [[VuoQtListener alloc] init];

	[listener initWithDevice:[NSString stringWithFormat:@"%s", inputDevice.name] id:[NSString stringWithFormat:@"%s", inputDevice.id] callback:receivedFrame];

	VuoRegister(listener, VuoQTCapture_free);

	return (VuoQTCapture) listener;
}

/**
 *	Change the device input.
 */
void VuoQtCapture_setInputDevice(VuoQTCapture movie, VuoVideoInputDevice inputDevice)
{
	VuoQtListener *listener = (VuoQtListener*)movie;
	[listener setInputDevice:[NSString stringWithFormat:@"%s", inputDevice.name] id:[NSString stringWithFormat:@"%s", inputDevice.id]];
}

/**
 * Begins receiving frames.
 */
void VuoQTCapture_startListening(VuoQTCapture movie)
{
	VuoQtListener *listener = (VuoQtListener*)movie;
	[listener startRunning];
}

/**
 * Stops receiving frames (starts ignoring/dropping them).
 */
void VuoQTCapture_stopListening(VuoQTCapture movie)
{
	VuoQtListener *listener = (VuoQtListener*)movie;
	[listener stopRunning];
}

/**
 *	Returns true if QT has been initialized, false otherwise.
 */
VuoBoolean VuoQTCapture_isInitialized(VuoQTCapture movie)
{
	if(movie == NULL)
		return false;

	VuoQtListener *listener = (VuoQtListener*)movie;

	return (bool)[listener isInitialized];
}

/**
 *	Set the image callback.
 */
void VuoQTCapture_setCallback(VuoQTCapture movie, VuoOutputTrigger(receivedFrame, VuoVideoFrame))
{
	VuoQtListener *listener = (VuoQtListener*)movie;
	[listener setCallback:receivedFrame];
}

/**
 * Destructor.
 */
void VuoQTCapture_free(VuoQTCapture movie)
{
	VuoQtListener *listener = (VuoQtListener *)movie;

	[listener stopRunning];
	[listener release];
}
